
;;;*********************************************************************
;;;
;;;               DonS-Ground<Prover.lisp    3 April 1989  26 May 1989
;;;                            
;;;       This is Don Simon's Matrix Prover for Ground Inequalities
;;;
;;;*********************************************************************

;;;Here's the program. I think we did prove it complete for the rationals (Don)

;;; New, Improved Ground Prover for Rational Inequalities.
;;; Written by Don Simon  11/2/87

;;; This file contains the code for a ground prover based on a matrix
;;; algorithm for finding contradictions and equalities from a set of
;;; rational inequalities.

;;; There are two ways to use this prover. The first is to give it the
;;; entire set of literals at once, by calling 
;;; (FIND-CONTRADICTION LITERALS). This function returns the values
;;; T NIL if a contradiction is found. Otherwise, the the values NIL
;;; INEQUALTIES are returned, where INEQUALITIES is a set of literals
;;; that are actually equalities.

;;; The second way to use the prover is to supply the literals one at
;;; time, and after each new literal is added, to ask if a contradiction
;;; or new equalities have been found. This could be done by calling
;;; FIND-CONTRADICTION with the augmented set of literals each time, but
;;; it's faster overall to save the intermediate work. This work is
;;; saved in a structure called an ineq-structure for lack of a better
;;; name. To do it this way, first call (MAKE-EMPTY-INEQ-STRUCTURE)
;;; which returns an empty ineq-structure. Then each time a new literal
;;; is to be added to the set, call 
;;; (ADD-LITERAL-TO-INEQ-STRUCTURE LITERAL INEQ-STRUCTURE) with the new
;;; literal and the ineq-structure. This function will return the values
;;; T NIL if a contradiction is found. Otherwise, the values NIL
;;; EQUALITIES are returned, where EQUALITIES is a set of literals
;;; that are should be equalities. The ineq-structure is changed by
;;; structure modification.

;;; (FIND-CONTRADICTION-ONLINE LITERALS) is an example of a function
;;; that finds a contradiction or equalities by adding the literals one
;;; at a time to the set.

;;; Both procedures expect the literals to be of the form (< A B) or
;;; (<= A B) where A and B are reduced. The function symbol "S*" should
;;; be used for scalar multiplication.

;;; ******************** Data Structures ********************

;;; ineq-structure	::	(subst-list . matrix*)
;;; matrix-list		::	(xxx . matrix*)
;;; subst-list		::	(maxvar . subst*)
;;; maxvar		::	integer
;;; subst		::	(term . integer)
;;; matrix		::	(first-column pos-set . neg-set)
;;; first-column	::	integer
;;; pos-set		::	iset
;;; neg-set	       	::	iset
;;; iset		::	(irow*)
;;; irow		::	(givens open-p offset . addend*)
;;; givens		::	(literal*)
;;; open-p		::	"t" | "nil"
;;; offset		::	number
;;; addend		::	(coefficient . multiplicand)
;;; coefficient		::	number
;;; multiplicand	::	integer

;;; ******************** Addends ********************

;;; addend		::	(coefficient . multiplicand)
;;; coefficient		::	number
;;; multiplicand	::	integer

;;; All expressions are represented as a sum of addends and an offset.
;;; All terms (variables and things headed by unevaluated function 
;;; symbols) have been replaced by integers so that equal terms
;;; are replaced by the same integer.

(DEFMACRO MAKE-ADDEND (COEFFICIENT MULTIPLICAND)
  "Makes an addend from a coefficient and multiplicand."
  `(CONS ,COEFFICIENT ,MULTIPLICAND))

(DEFMACRO COEFFICIENT (ADDEND)
  "Returns the coefficient of an addend."
  `(CAR ,ADDEND))

(DEFMACRO MULTIPLICAND (ADDEND)
  "Returns the multiplicand of an addend."
  `(CDR ,ADDEND))

(DEFMACRO MULTIPLY-ADDEND (MULT ADDEND)
  "Multiplies addend by the number mult."
  `(MAKE-ADDEND (* ,MULT (COEFFICIENT ,ADDEND)) (MULTIPLICAND ,ADDEND)))

(DEFMACRO SORT-ADDENDS (ADDENDS)
  "Sort a list of addends by their multiplicands."
  `(SORT ,ADDENDS #'< :KEY #'CDR))

(DEFUN MULTIPLY-AND-ADD-ADDENDS (MULT ADDENDS-1 ADDENDS-2)
  "Multiplies ADDENDS-2 by MULT and adds to ADDENDS-1. Addends must be sorted."
  (LET* ((NEW-ADDENDS (LIST NIL))
	 (TAIL NEW-ADDENDS))
    (DO ()
	((COND ((NULL ADDENDS-2) (SETF (CDR TAIL) ADDENDS-1) T)
	       ((NULL ADDENDS-1) 
		(SETF (CDR TAIL) 
		      (LOOP FOR ADDEND-2 IN ADDENDS-2 
			    COLLECTING (MULTIPLY-ADDEND MULT ADDEND-2)))
		T))
	 (CDR NEW-ADDENDS))
      (LET ((ADDEND-1 (FIRST ADDENDS-1))
	    (ADDEND-2 (FIRST ADDENDS-2)))
	(COND ((EQ (MULTIPLICAND ADDEND-1) (MULTIPLICAND ADDEND-2))
	       (LET ((NEW-COEFFICIENT (+ (COEFFICIENT ADDEND-1) (* MULT (COEFFICIENT ADDEND-2)))))
		 (UNLESS (ZEROP NEW-COEFFICIENT)
		   (SETF (CDR TAIL)
			 (SETQ TAIL (LIST (MAKE-ADDEND NEW-COEFFICIENT (MULTIPLICAND ADDEND-1))))))
		 (SETQ ADDENDS-1 (CDR ADDENDS-1))
		 (SETQ ADDENDS-2 (CDR ADDENDS-2))))
	      ((< (MULTIPLICAND ADDEND-1) (MULTIPLICAND ADDEND-2))
	       (SETF (CDR TAIL)
		     (SETQ TAIL (LIST ADDEND-1)))
	       (SETQ ADDENDS-1 (CDR ADDENDS-1)))
	      (T (SETF (CDR TAIL)
		       (SETQ TAIL (LIST (MULTIPLY-ADDEND MULT ADDEND-2))))
		 (SETQ ADDENDS-2 (CDR ADDENDS-2))))))))

;;; ******************** Irow ********************

;;; irow		::	(givens open-p offset . addend*)

;;; All inequalities are represented by irows. The givens are the literals
;;; from the original set that prove the inequality. The inequality represented
;;; by the irow (givens T n (coef1 . mult1) ... (coefm . multm)) is
;;; (<= 0 (+ n (* coef1 term_mult1) ... (* coefm term_multm)). 
;;; The inequality represented by the irow 
;;; (givens NIL n (coef1 . mult1) ... (coefm . multm)) is
;;; (< 0 (+ n (* coef1 term_mult1) ... (* coefm term_multm)).

(DEFMACRO MAKE-IROW (GIVENS OPEN-P OFFSET ADDENDS)
  "Makes an irow from the givens, open-p, offset, and addends."
  `(LIST* ,GIVENS ,OPEN-P ,OFFSET ,ADDENDS))

(DEFMACRO GIVENS (IROW)
  "Returns the givens of an irow."
  `(CAR ,IROW))

(DEFMACRO OPEN-P (IROW)
  "Does irow represent a non-strict inequality?"
  `(CADR ,IROW))

(DEFMACRO OFFSET (IROW)
  "Returns the offset of an irow."
  `(CADDR ,IROW))

(DEFMACRO ADDENDS (IROW)
  "Returns the addends of a irow."
  `(CDDDR ,IROW))

(DEFMACRO NEGATIVE-ROW-P (IROW)
  "Is the coefficient of the first addend of an irow negative?"
  `(MINUSP (COEFFICIENT (FIRST (ADDENDS ,IROW)))))

(DEFMACRO CONTRADICTORY-ROW-P (IROW)
  "Does irow represent a contradiction?"
  ;; Assumes that IROW has no addends.
  `(OR (MINUSP (OFFSET ,IROW))
       (AND (ZEROP (OFFSET ,IROW)) (NOT (OPEN-P ,IROW)))))

(DEFMACRO EQUALITY-ROW-P (IROW)
  "Assuming that irow is not a contradiction, must it be an equality?"
  ;; Assumes that IROW has no addends and is not contradictory.
  `(ZEROP (OFFSET ,IROW)))

;;; ******************** Subst-list ********************

;;; subst-list		::	(maxvar . subst*)
;;; maxvar		::	integer
;;; subst		::	(term . integer)

;;; The subst-list is used to replace the terms of an expression
;;; by integers. It is just maxvar consed onto a substitution
;;; list. Maxvar is the largest of all the right-hand sides of
;;; the substitutions.

(DEFMACRO MAKE-SUBST-LIST ()
  "Makes an empty subst-list."
  `(LIST 0))

(DEFMACRO MAXVAR (SUBST-LIST)
  "Returns the largest right-hand side in the subst-list."
  `(CAR ,SUBST-LIST))

(DEFMACRO SUBSTS (SUBST-LIST)
  "Returns the substs (substitution list) of a subst-list."
  `(CDR ,SUBST-LIST))

;;; ******************** Ineq-structure ********************

;;; ineq-structure	::	(subst-list . matrix*)
;;; matrix-list		::	(xxx . matrix*)
;;; matrix		::	(first-column pos-set . neg-set)
;;; first-column	::	integer
;;; pos-set		::	iset
;;; neg-set	       	::	iset

;;; An ineq-structure is comprised of two parts: The subst-list used
;;; to transform the literals into irows, and a list of the matrices
;;; which are the intermediate results of the search for a contradiction.
;;; Each matrix is represented by the index of the first non-zero column
;;; (first-column), a set of irows whose coefficient in that column is
;;; positive (pos-set), and a set of irows whose coefficient in that
;;; column is negative (neg-set). The list of matrices is ordered by
;;; first-column.

;;; a matrix-list is anything consed onto a list of matrices. The extra
;;; cell is there so a new matrix can be added to the head of the list
;;; using structure modification.

(DEFMACRO MAKE-INEQ-STRUCTURE (SUBST-LIST MATRICES)
  "Makes an ineq-structure from a subst-list and a list of matrices."
  `(CONS ,SUBST-LIST ,MATRICES))

(DEFMACRO MAKE-EMPTY-INEQ-STRUCTURE ()
  "Makes an empty ineq-structure."
  `(CONS (MAKE-SUBST-LIST) NIL))

(DEFMACRO SUBST-LIST (INEQ-STRUCTURE)
  "Returns the subst-list of an ineq-structure."
  `(CAR ,INEQ-STRUCTURE))

(DEFMACRO MATRICES (INEQ-STRUCTURE)
  "Returns the matrices of an ineq-structure."
  `(CDR ,INEQ-STRUCTURE))

(DEFMACRO MAKE-MATRIX (FIRST-COLUMN POS-SET NEG-SET)
  "Makes a matrix from the first-column, pos-set and neg-set."
  `(LIST* ,FIRST-COLUMN ,POS-SET ,NEG-SET))

(DEFMACRO FIRST-COLUMN (MATRIX)
  "Returns the index of the first non-zero column of a matrix."
  `(CAR ,MATRIX))

(DEFMACRO POS-SET (MATRIX)
  "Returns the set of irows whose first non-zero coefficient is positive."
  `(CADR ,MATRIX))

(DEFMACRO NEG-SET (MATRIX)
  "Returns the set of irows whose first non-zero coefficient is negative."
  `(CDDR ,MATRIX))

;;; ******************** Conversion Routines ********************

;;; These routines convert a literal to an irow by replacing the terms
;;; with integers, and putting the resulting expression into standard
;;; form. The standard form is that all expressions are sums which
;;; always have a numeric offset (possibly zero) and a list of addends
;;; which are always scalar products (whose coefficient may be 1). Thus,
;;; the expression is represented by the offset and the list of addends
;;; of the form (coefficient . multiplicand). The list of addends is
;;; ordered by the multiplicands. The left-hand side of the inequality
;;; is subtracted from the right-hand side. Open-p is T iff the
;;; inequality was non-strict. The givens of an irow is just the list
;;; containing the original literal.

;;; Note that SUBST-LIST will be changed by structure modification in
;;; these procedures. 

(DEFUN CONVERT-LITERAL (LITERAL SUBST-LIST)
  "Converts a literal to an irow using subst-list. Subst-list may be modified."
  (MULTIPLE-VALUE-BIND (OFFSET-1 ADDENDS-1)
      (CONVERT-EXPR (SECOND LITERAL) SUBST-LIST)
    (MULTIPLE-VALUE-BIND (OFFSET-2 ADDENDS-2)
	(CONVERT-EXPR (THIRD LITERAL) SUBST-LIST)
      (MAKE-IROW (LIST LITERAL)
		 (EQ (FIRST LITERAL) '<=)
		 (- OFFSET-2 OFFSET-1)
		 (MULTIPLY-AND-ADD-ADDENDS -1 ADDENDS-2 ADDENDS-1)))))

(DEFUN CONVERT-EXPR (EXPR SUBST-LIST)
  "Replaces the terms of expr by integers using subst-list and converts the result to a standard form.
   Returns the values OFFSET ADDENDS. Subst-list may be modified."
  (COND ((NUMBERP EXPR) (VALUES EXPR NIL))
	((AND (LISTP EXPR) (EQ (FIRST EXPR) '+))
	 (LOOP FOR A IN (CDR EXPR)
	       IF (NUMBERP A) SUM A INTO OFFSET
		 ELSE COLLECT (CONVERT-ADDEND A SUBST-LIST) INTO ADDENDS
	       FINALLY (RETURN (VALUES OFFSET (SORT-ADDENDS ADDENDS)))))
;	 (IF (NUMBERP (SECOND EXPR))
;	     (VALUES (SECOND EXPR)
;		     (SORT-ADDENDS
;		       (LOOP FOR ADDEND IN (CDDR EXPR) 
;			   COLLECTING (CONVERT-ADDEND ADDEND SUBST-LIST))))
;	     (VALUES 0
;		     (SORT-ADDENDS
;		       (LOOP FOR ADDEND IN (CDR EXPR) 
;			     COLLECTING (CONVERT-ADDEND ADDEND SUBST-LIST))))))
	(T (VALUES 0 (LIST (CONVERT-ADDEND EXPR SUBST-LIST))))))

(DEFUN CONVERT-ADDEND (ADDEND SUBST-LIST)
  "Converts a scalar product or term to an addend using subst-list. Subst-list may be modified."
  (IF (AND (LISTP ADDEND) (EQ (FIRST ADDEND) 'S*))
      (MAKE-ADDEND (SECOND ADDEND) (CONVERT-TERM (THIRD ADDEND) SUBST-LIST))
      (MAKE-ADDEND 1 (CONVERT-TERM ADDEND SUBST-LIST))))

(DEFUN CONVERT-TERM (TERM SUBST-LIST)
  "Converts a term to its equivalent integer using subst-list. Subst-list may be modified."
  (LET ((REPLACEMENT (ASSOC TERM (CDR SUBST-LIST) :TEST #'EQUAL)))
    (COND ((NULL REPLACEMENT)
	   (LET ((VAR-NUMBER (INCF (MAXVAR SUBST-LIST))))
	     (SETF (SUBSTS SUBST-LIST) (CONS (CONS TERM VAR-NUMBER) (SUBSTS SUBST-LIST)))
	      VAR-NUMBER))
	  (T (CDR REPLACEMENT)))))

;;; ******************** Contradiction ********************

(DEFUN ADD-LITERAL-TO-INEQ-STRUCTURE (LITERAL INEQ-STRUCTURE)
  "Adds a literal to an ineq-structure. Returns the values T NIL if a contradiction is found.
   Otherwise, it returns the values NIL EQUALITIES. INEQ-STRUCTURE is changed by structure modification."
  (LET ((IROW (CONVERT-LITERAL LITERAL (SUBST-LIST INEQ-STRUCTURE))))
    (COND ((NOT (NULL (ADDENDS IROW)))
	   (ADD-IROW-TO-MATRIX-LIST
	     IROW
	     (MULTIPLICAND (FIRST (ADDENDS IROW)))
	     INEQ-STRUCTURE))
	  ((CONTRADICTORY-ROW-P IROW) (VALUES T NIL))
	  ((EQUALITY-ROW-P IROW) (VALUES NIL (GIVENS IROW)))
	  (T (VALUES NIL NIL)))))

(DEFUN ADD-IROW-TO-MATRIX-LIST (IROW MINVAR MATRIX-LIST)
  "Adds an irow to a matrix-list. Returns the values T NIL if a contradiction is found.
   Otherwise, it returns the values NIL EQUALITIES. IROW must have addends and minvar
   is the smallest multiplicand in IROW. MATRIX-LIST is changed by structure modification."
  (COND ((OR (NULL (CDR MATRIX-LIST)) (< MINVAR (FIRST-COLUMN (SECOND MATRIX-LIST))))
	 (SETF (CDR MATRIX-LIST)
	       (CONS (IF (NEGATIVE-ROW-P IROW)
			 (MAKE-MATRIX MINVAR NIL (LIST IROW))
			 (MAKE-MATRIX MINVAR (LIST IROW) NIL))
		     (CDR MATRIX-LIST)))
	 (VALUES NIL NIL))
	((> MINVAR (FIRST-COLUMN (SECOND MATRIX-LIST)))
	 (ADD-IROW-TO-MATRIX-LIST IROW MINVAR (CDR MATRIX-LIST)))
	((NEGATIVE-ROW-P IROW)
	 (LOOP WITH EQUALITIES = NIL 
	       FOR POS-ROW IN (POS-SET (SECOND MATRIX-LIST))
	       FOR NEW-ROW = (COMBINE-IROWS POS-ROW IROW)
	       IF (NOT (NULL (ADDENDS NEW-ROW))) 
		 MINIMIZE (MULTIPLICAND (FIRST (ADDENDS NEW-ROW))) INTO NEW-MINVAR
		 AND COLLECT NEW-ROW INTO ZERO-SET
	       ELSE IF (CONTRADICTORY-ROW-P NEW-ROW) RETURN (VALUES T NIL)
	       ELSE IF (EQUALITY-ROW-P NEW-ROW) 
		      DO (SETQ EQUALITIES (UNION EQUALITIES (GIVENS NEW-ROW)))
	       FINALLY (RETURN
			 (MULTIPLE-VALUE-BIND (CONTRADICTION NEW-EQUALITIES)
			     (ADD-ISET-TO-MATRIX-LIST ZERO-SET NEW-MINVAR (CDR MATRIX-LIST) EQUALITIES)
			   (COND (CONTRADICTION (VALUES T NIL))
				 (T (SETF (NEG-SET (SECOND MATRIX-LIST)) 
					  (CONS IROW (NEG-SET (SECOND MATRIX-LIST))))
				    (VALUES NIL NEW-EQUALITIES)))))))
	(T (LOOP WITH EQUALITIES = NIL 
		 FOR NEG-ROW IN (NEG-SET (SECOND MATRIX-LIST))
		 FOR NEW-ROW = (COMBINE-IROWS IROW NEG-ROW)
		 IF (NOT (NULL (ADDENDS NEW-ROW))) 
		   MINIMIZE (MULTIPLICAND (FIRST (ADDENDS NEW-ROW))) INTO NEW-MINVAR
		   AND COLLECT NEW-ROW INTO ZERO-SET
		 ELSE IF (CONTRADICTORY-ROW-P NEW-ROW) RETURN (VALUES T NIL)
		 ELSE IF (EQUALITY-ROW-P NEW-ROW) 
			DO (SETQ EQUALITIES (UNION EQUALITIES (GIVENS NEW-ROW)))
		 FINALLY (RETURN
			   (MULTIPLE-VALUE-BIND (CONTRADICTION NEW-EQUALITIES)
			       (ADD-ISET-TO-MATRIX-LIST ZERO-SET NEW-MINVAR (CDR MATRIX-LIST) EQUALITIES)
			     (COND (CONTRADICTION (VALUES T NIL))
				   (T (SETF (POS-SET (SECOND MATRIX-LIST))
					    (CONS IROW (POS-SET (SECOND MATRIX-LIST))))
				      (VALUES NIL NEW-EQUALITIES)))))))))

(DEFUN ADD-ISET-TO-MATRIX-LIST (ISET MINVAR MATRIX-LIST EQUALITIES)
  "Adds a set of irows to a matrix-list. Returns the values T NIL if a contradiction is found.
   Otherwise, it returns the values NIL NEW-EQUALITIES. MINVAR is the index of the first non-zero column in ISET.
   EQUALITIES is the current set of equalities. MATRIX-LIST is changed by structure modification."
  (COND ((NULL ISET) (VALUES NIL EQUALITIES))
	((NULL (CDR MATRIX-LIST)) 
	 (MULTIPLE-VALUE-BIND (CONTRADICTION EQUALITIES MATRICES)
	     (FIND-CONTRADICTION-IN-ISET ISET MINVAR EQUALITIES)
	   (COND (CONTRADICTION (VALUES T NIL))
		 (T (SETF (CDR MATRIX-LIST) MATRICES)
		    (VALUES NIL EQUALITIES)))))
	((< MINVAR (FIRST-COLUMN (SECOND MATRIX-LIST)))
	 (MULTIPLE-VALUE-BIND (POS-SET NEG-SET ZERO-SET NEW-MINVAR)
	     (SPLIT-ISET ISET MINVAR)
	   (MULTIPLE-VALUE-BIND (CONTRADICTION ZERO-SET-1 MINVAR-1 EQUALITIES-1)
	       (COMBINE-ISETS POS-SET NEG-SET ZERO-SET NEW-MINVAR EQUALITIES)
	     (IF CONTRADICTION
		 (VALUES T NIL)
		 (MULTIPLE-VALUE-BIND (CONTRADICTION NEW-EQUALITIES)
		     (ADD-ISET-TO-MATRIX-LIST ZERO-SET-1 MINVAR-1 MATRIX-LIST EQUALITIES-1)
		   (COND (CONTRADICTION (VALUES T NIL))
			 (T (SETF (CDR MATRIX-LIST)
				  (CONS (MAKE-MATRIX MINVAR POS-SET NEG-SET)
					(CDR MATRIX-LIST)))
			    (VALUES NIL NEW-EQUALITIES))))))))
	((> MINVAR (FIRST-COLUMN (SECOND MATRIX-LIST)))
	 (ADD-ISET-TO-MATRIX-LIST ISET MINVAR (CDR MATRIX-LIST) EQUALITIES))
	(T (MULTIPLE-VALUE-BIND (POS-SET NEG-SET ZERO-SET NEW-MINVAR)
	       (SPLIT-ISET ISET MINVAR)
	     (MULTIPLE-VALUE-BIND (CONTRADICTION ZERO-SET-1 MINVAR-1 EQUALITIES-1)
		 (COMBINE-ISETS POS-SET NEG-SET ZERO-SET NEW-MINVAR EQUALITIES)
	       (IF CONTRADICTION
		   (VALUES T NIL)
		   (MULTIPLE-VALUE-BIND (CONTRADICTION ZERO-SET-2 MINVAR-2 EQUALITIES-2)
		       (COMBINE-ISETS POS-SET (NEG-SET (SECOND MATRIX-LIST)) ZERO-SET-1 MINVAR-1 EQUALITIES-1)
		     (IF CONTRADICTION
			 (VALUES T NIL)
			 (MULTIPLE-VALUE-BIND (CONTRADICTION ZERO-SET-3 MINVAR-3 EQUALITIES-3)
			     (COMBINE-ISETS (POS-SET (SECOND MATRIX-LIST)) NEG-SET ZERO-SET-2 MINVAR-2 EQUALITIES-2)
			   (IF CONTRADICTION
			       (VALUES T NIL)
			       (MULTIPLE-VALUE-BIND (CONTRADICTION NEW-EQUALITIES)
				   (ADD-ISET-TO-MATRIX-LIST ZERO-SET-3 MINVAR-3 (CDR MATRIX-LIST) EQUALITIES-3)
				 (COND (CONTRADICTION (VALUES T NIL))
				       (T (SETF (POS-SET (SECOND MATRIX-LIST))
						(APPEND POS-SET (POS-SET (SECOND MATRIX-LIST))))
					  (SETF (NEG-SET (SECOND MATRIX-LIST))
						(APPEND NEG-SET (NEG-SET (SECOND MATRIX-LIST))))
					  (VALUES NIL NEW-EQUALITIES))))))))))))))

(DEFUN FIND-CONTRADICTION-IN-ISET (ISET MINVAR EQUALITIES)
  "Tries to find a contradiction in non-null ISET whose first non-zero column index is MINVAR. Returns T NIL NIL
   if a contradiction is found. Otherwise, returns NIL NEW-EQUALITIES MATRICES where MATRICES is
   the list of intermediate matrices. EQUALITIES is the current set of equalities."
  ;; ISET is non-null.
  (MULTIPLE-VALUE-BIND (POS-SET NEG-SET ZERO-SET NEW-MINVAR)
      (SPLIT-ISET ISET MINVAR)
    (MULTIPLE-VALUE-BIND (CONTRADICTION ZERO-SET-1 MINVAR-1 EQUALITIES-1)
	(COMBINE-ISETS POS-SET NEG-SET ZERO-SET NEW-MINVAR EQUALITIES)
      (COND (CONTRADICTION (VALUES T NIL NIL))
	    ((NULL ZERO-SET-1) 
	     (VALUES NIL EQUALITIES (LIST (MAKE-MATRIX MINVAR POS-SET NEG-SET))))
	    (T (MULTIPLE-VALUE-BIND (CONTRADICTION NEW-EQUALITIES NEW-MATRICES)
		   (FIND-CONTRADICTION-IN-ISET ZERO-SET-1 MINVAR-1 EQUALITIES-1)
		 (IF CONTRADICTION
		     (VALUES T NIL NIL)
		     (VALUES
		       NIL
		       NEW-EQUALITIES
		       (CONS (MAKE-MATRIX MINVAR POS-SET NEG-SET) NEW-MATRICES)))))))))

(DEFUN COMBINE-ISETS (POS-SET NEG-SET ZERO-SET NEW-MINVAR EQUALITIES)
  "Adds pos-set to neg-set and appends the result to zero-set. Returns the values T if
   a contradiction is found. Otherwise, returns NIL NEW-ZERO-SET NEWEST-MINVAR NEW-EQUALITIES.
   NEW-MINVAR is the index of the first non-zero column in ZERO-SET. EQUALITIES is the current set of equalities."
  (LOOP FOR POS-ROW IN POS-SET
	THEREIS (LOOP FOR NEG-ROW IN NEG-SET
		      FOR NEW-ROW = (COMBINE-IROWS POS-ROW NEG-ROW)
		      IF (NOT (NULL (ADDENDS NEW-ROW)))
			DO (SETQ NEW-MINVAR 
				 (MIN NEW-MINVAR (MULTIPLICAND (FIRST (ADDENDS NEW-ROW)))))
			AND DO (SETQ ZERO-SET (CONS NEW-ROW ZERO-SET))
		      ELSE IF (CONTRADICTORY-ROW-P NEW-ROW) RETURN T
		      ELSE IF (EQUALITY-ROW-P NEW-ROW) 
			     DO (SETQ EQUALITIES (UNION EQUALITIES (GIVENS NEW-ROW))))
	FINALLY (RETURN (VALUES NIL ZERO-SET NEW-MINVAR EQUALITIES))))

(DEFUN SPLIT-ISET (ISET MINVAR)
  "Splits ISET into POS-SET, NEG-SET, and ZERO-SET. Returns these and the index of the first non-zero
   column in ZERO-SET."
  (LOOP WITH NEW-MINVAR = 1000
	FOR IROW IN ISET
	FOR ADDEND = (FIRST (ADDENDS IROW))
	IF (> (MULTIPLICAND ADDEND) MINVAR) 
	  DO (SETQ NEW-MINVAR (MIN NEW-MINVAR (MULTIPLICAND ADDEND)))
	  AND COLLECT IROW INTO ZERO-SET
	ELSE IF (MINUSP (COEFFICIENT ADDEND)) COLLECT IROW INTO NEG-SET
	ELSE COLLECT IROW INTO POS-SET
	FINALLY (RETURN (VALUES POS-SET NEG-SET ZERO-SET NEW-MINVAR))))

(DEFUN COMBINE-IROWS (POS-ROW NEG-ROW)
  "Adds a positive and negative row."
  (LET ((MULT (/ (COEFFICIENT (FIRST (ADDENDS POS-ROW))) 
		 (- (COEFFICIENT (FIRST (ADDENDS NEG-ROW)))))))
    (MAKE-IROW (UNION (GIVENS POS-ROW) (GIVENS NEG-ROW))
	       (AND (OPEN-P POS-ROW) (OPEN-P NEG-ROW))
	       (+ (OFFSET POS-ROW) (* MULT (OFFSET NEG-ROW)))
	       (MULTIPLY-AND-ADD-ADDENDS MULT 
					 (CDR (ADDENDS POS-ROW))
					 (CDR (ADDENDS NEG-ROW))))))

(DEFUN FIND-CONTRADICTION-ONLINE (LITERALS)
  "Sample function to show how a contradiction is found taking literals one at a time."
  (LOOP WITH INEQ-STRUCTURE = (MAKE-EMPTY-INEQ-STRUCTURE)
	FOR LITERAL IN LITERALS
	FOR (CONTRADICTION EQUALITIES)
	    = (MULTIPLE-VALUE-LIST (ADD-LITERAL-TO-INEQ-STRUCTURE LITERAL INEQ-STRUCTURE))
	IF CONTRADICTION RETURN (VALUES T NIL)
	APPENDING EQUALITIES INTO ALL-EQUALITIES
	FINALLY (RETURN (VALUES NIL ALL-EQUALITIES))))

(DEFUN FIND-CONTRADICTION (LITERALS)
  "Finds a contradiction in the set of literals. Returns the values T NIL if a contradiction is found.
   Otherwise, returns the values NIL EQUALITIES."
 (block Main
   (LOOP 
	WITH MINVAR = 1 AND EQUALITIES = NIL AND SUBST-LIST = (MAKE-SUBST-LIST)
	FOR LITERAL IN LITERALS
	FOR IROW = (CONVERT-LITERAL LITERAL SUBST-LIST)
	IF (NOT (NULL (ADDENDS IROW))) COLLECT IROW INTO ISET
	ELSE IF (CONTRADICTORY-ROW-P IROW) RETURN (VALUES T NIL)
	     ELSE IF (EQUALITY-ROW-P IROW) DO (SETQ EQUALITIES (UNION EQUALITIES (GIVENS IROW)))
	FINALLY (LOOP UNTIL (NULL ISET)
		      DO (LOOP WITH NEW-MINVAR = (FIRST SUBST-LIST)
			       FOR IROW IN ISET
			       FOR ADDEND = (FIRST (ADDENDS IROW))
			       IF (> (MULTIPLICAND ADDEND) MINVAR) 
				 DO (SETQ NEW-MINVAR (MIN NEW-MINVAR (MULTIPLICAND ADDEND)))
				 AND COLLECT IROW INTO NEW-SET
			       ELSE IF (MINUSP (COEFFICIENT ADDEND)) COLLECT IROW INTO NEG-SET
				    ELSE COLLECT IROW INTO POS-SET
			       FINALLY (LOOP FOR POS-ROW IN POS-SET
					     DO (LOOP FOR NEG-ROW IN NEG-SET
						      FOR MULT = (/ (COEFFICIENT (FIRST (ADDENDS POS-ROW))) 
								    (- (COEFFICIENT (FIRST (ADDENDS NEG-ROW)))))
						      FOR NEW-ROW = (MAKE-IROW (UNION (GIVENS POS-ROW) (GIVENS NEG-ROW))
									       (AND (OPEN-P POS-ROW) (OPEN-P NEG-ROW))
									       (+ (OFFSET POS-ROW) (* MULT (OFFSET NEG-ROW)))
									       (MULTIPLY-AND-ADD-ADDENDS MULT 
													 (CDR (ADDENDS POS-ROW))
													 (CDR (ADDENDS NEG-ROW))))
						      IF (NOT (NULL (ADDENDS NEW-ROW)))
							DO (SETQ NEW-MINVAR 
								 (MIN NEW-MINVAR (MULTIPLICAND (FIRST (ADDENDS NEW-ROW)))))
							AND DO (SETQ NEW-SET (CONS NEW-ROW NEW-SET))
						      ELSE IF (CONTRADICTORY-ROW-P NEW-ROW) DO (RETURN-FROM MAIN (VALUES T NIL))
							   ELSE IF (EQUALITY-ROW-P NEW-ROW) 
								  DO (SETQ EQUALITIES (UNION EQUALITIES (GIVENS NEW-ROW))))
					     FINALLY (SETQ ISET NEW-SET
							   MINVAR NEW-MINVAR))))
		(RETURN (VALUES NIL EQUALITIES)))))

