
;;;****************************************************************************
;;;
;;;                  VE.LISP                 29 Aug 1987   16 March 1988
;;;                                                         2 May   1989
;;;
;;;     This is an package for handling INEQUALITY proofs.  It contains
;;;
;;;        *  a VE package for doing Variable Elimination and Shielding-term
;;;             removal
;;;
;;;        *  a GROUND Inequality Prover  
;;;
;;;----------------------------------------------------------------------------
;;;
;;;     The Main function here is INEQUALITY-PROVE. It is called from OPTIONS
;;;     or from  IMPLY-RULES*.  These are  part of the IMPLY.LISP package,
;;;     our Natural Deduction Prover, based on the YET principle.
;;;
;;;     From options, INEQUALITY-PROVE is called by the user command INEQ
;;;     From imply-rules*, it is called if the *VE-light* is not nill, and
;;;     a) C is a pure-inequality formula, or
;;;     b) C is of the form (< a b) or (<= a b).
;;
;;;     INIQUALITY-PROVE accomplishes the proofs using  a) the ground prover, 
;;;     b) VE (varialbe elimination)  or c) Shielding Term removal.
;;;     (See CADE-1980 paper).
;;;
;;;     For this purpose, it has access not only to the goal  C, but
;;;     also to the hypothesis,  H,  and YET, the list of
;;;     (hypothesis, Goal) pairs (and supporting data) which have yet
;;;     to be proved before the whole proof is completed.  E.g., in order
;;;     to use VE on a variable x, that variable must be eligible in
;;;     all of  C ,  H , and YET.
;;;
;;;****************************************************************************

(defun inequality-prove (h c tl anc lt yet)     ;called in IMPLY-RULES
  ;; Uses VE, Ground-prover, FLIPPING, and Shielding-term removal methods
  ;; of the Bledsoe & Hines CADE 6 Paper.
  ;; Recall that h = (use h-gu h= hv). This function revised 2 June 89
  ;; We assume that it always finishes with a successful call to GROUND-PROVER.

  (declare (special h c tl anc lt yet))

  (let (hv vc ec ec1 ec2 ec3 sc vh eh eh1 eh2 eh3 vy sy)
  
    (setq hv (fourth h))
    (setq vy (crunch (mapcan #'(lambda (u) 
	(append (free-variables-in (cadr u)) (free-variables-in (caddr u))))
			     yet)))
    (setq sy (crunch (mapcan #'(lambda (u) 
	(append (shieldeds (cadr u) vy) (shieldeds (caddr u) vy)))
			     yet)))
    (cond 
     ((setq vc (free-variables-in c))
      (cond
       ((setq ec (eligibles c vc))
	(cond ((setq ec1 (set-difference ec vy))
	       (VE-the-variable (car ec1) t))  ;VE without using H and YET
	      ((setq ec2 (set-difference (set-difference ec (shieldeds hv ec))
					(shieldeds-yet yet ec)))
	       (VE-the-variable (car ec1) nil))  ;VE using H and YET
	      ((setq vh (free-variables-in hv))
	       (cond ((setq eh (eligibles hv vy))
		      (cond ((setq eh1 (set-difference eh vy)) (flip-h eh1))
			     ;;move a conjunct containing x from h to C to VE x
			    ((setq eh2 (set-difference (set-difference eh
                                (setq sc (shieldeds c eh)))
				(shieldeds-yet yet eh))) (flip-h eh2))    
			     ;;move a conjunct containing x from h to C to VE x
			    ((setq eh3 (set-intersection eh sc))
			     (remove-shield c eh3))
			    ((setq ec3 (set-intersection ec (shieldeds h ec)))
			     (flip-h (car ec3) t)) 
			     ;;Move a shielding term from h to C to STR it
			    (t (flip-yet (set-intersection ec sy) t))))
			     ;;Move a shielding term from yet to C to STR it
		     ((setq ec3 (set-intersection ec (shieldeds h ec)))
		      (flip-h ec3) t)
                       ;;Move a shielding term from h to C to STR it
		     (t (flip-yet (set-intersection ec sy) t))))
                       ;;Move a shielding term from yet to C to STR it	  
	      (t (flip-yet (set-intersection ec sy) t))))
       
       ((setq vh (free-variables-in hv))
	(cond ((setq eh (eligibles hv vy))
	       (cond ((setq eh1 (set-difference eh vy)) (flip-h eh1))
		     ;;move a conjunct containing x from h to C to VE x
		     ((setq eh2 (set-difference (set-difference eh
	  	        	(setq sc (shieldeds c eh)))
				(shieldeds-yet yet eh))) (flip-h eh2))    
		     ;;move a conjunct containing x from h to C to VE x
		     ((setq eh3 (set-intersection eh sc))
		      (remove-shield c eh3))
		     (t (flip-yet (set-intersection ec sy) t))))
	              ;;Move a shielding term from yet to C to STR it
	      (t (remove-shield c (shieldeds c vc)))))
       (t (remove-shield c (shieldeds c vc)))))

     ((ground-prover (imp-on (second H) C))
      ;;try the ground prover, since C is ground.
      (format t "~%The GROUND-PROVER proved C= ~%       ~s~%~%" 
	      (imp-on (second H) C))
      (prove-yet t))			

     ((setq vh (free-variables-in hv))
	(cond ((setq eh (eligibles hv vy))
	       (cond ((setq eh1 (set-difference eh vy)) (flip-h eh1))
		     ;;move a conjunct containing x from h to C to VE x
		     ((setq eh2 (set-difference eh (shieldeds-yet yet eh)))
		      (flip-h eh2))    
		     ;;move a conjunct containing x from h to C to VE x
		     ((setq eh3 (set-intersection eh sc))
		      (remove-shield c eh3))
		     (t (flip-yet (set-intersection ec sy) t))))
	              ;;Move a shielding term from yet to C to STR it
	      (t (flip-h (shieldeds h vh) t))))

     (t nil) )))



(defun andify (l)			;used in VE
  ;;convert a list into a conjunction
  (cond ((null l) nil)
	((null (cdr l)) (car l))
        (t (cons '& l))))

(defun &H (h)  ;called by options
  ;;If H has the form (use H-gu Hv H=) then return (& use H-gu Hv H=)
  ;;  else return h
  (cond ((atom h) h)
	((not (length h 4)) h)
	((atom (car h)) h)
	(t (cons '& h))))

(defun and-or-together (L)
  ;;eg, ((a b)(c d)) goes to  (& (or a c) (or a d) (or b c) (or b e))
  (cons '& (mapcar #'(lambda (u) (cons 'or u)) (a-o-t L))))

(defun or-and-together (L)
  ;;eg, ((a b)(c d)) goes to  (or (& a c) (& a d) (& b c) (& b e))
  (cons 'or (mapcar #'(lambda (u) (cons '& u)) (a-o-t L))))

(defun a-o-t (L)        ;Called by and-or-together, or-and-together
  ;;eg, ((a b)(c d)(e f)) goes to ((a c e)(a c f)(a d e)(a d f)...(b d f))
  (let (L*)
   (cond
    ((null L) nil)
    ((null (cdr L)) (mapcar #'list (car L)))   ;;((e f)) --> ((e)(f))
    (t (setq L* (a-o-t (cdr L)))
       (mapcan #'(lambda (x) (mapcar #'(lambda (y)(cons x y)) L*)) (car L))))))

(defun clear-& (L)     ;called by VE
  ;; This converts (& B) to B, (&) to T, leaves (& B C) alone, etc
  (cond ((null (cdr L)) T) ((null (cddr L)) (cadr L)) (T L)))

(defun clear-or (L)     ;called by VE
  ;; This converts (or B) to B, (or) to T, leaves (or B C) alone, etc
  (cond ((null (cdr L)) T) ((null (cddr L)) (cadr L)) (T L)))

(defun clear-+ (L)     ;called by PLUS-ON
  ;; This converts (+ B) to B, (+) to 0, leaves (+ B C) alone, etc
  (cond ((atom L) L) ((not (eq '+ (car L))) L)
	((null (cdr L)) 0) ((null (cddr L)) (cadr L)) (T L)))

(defun conjunctive-unit (L x)    ;called by VE
  ;; Sellect a member of L which contains x and is of the form (& a ...)

 (cond ((null L) nil)
       ((not (occur x L)) nil)
       ((eq '& (caar L)) (car L))
       (t (conjunctive-unit (cdr L) x)) ))

(defun constant-part< (A x &optional Ac Ax)  ;called by VE   26Feb90
 ;;A is a list.  This returns a pair W = (Ac Ax) where A = Ac U Ax,
 ;;  members of Ac do not contain x, and members of Ax contain x and 
 ;;  are inequality atoms (of the form (< ...) or (<= ...)), if that
 ;;  is possible.  Else it returns NIL.

  (cond ((null A) (list Ac Ax))
	((not (occur x (car A))) 
	 (constant-part< (cdr A) x (cons (car A) Ac) Ax))
	((atom (car A)) nil)
	((member (caar A) '(< <=))
	 (constant-part< (cdr A) x Ac (cons (car A) Ax)))
	(t nil)))

(defun disjunct-conjunct-unit (A x)
  ;;A has the form (& ...), or (or ...).
  ;;This returns a list (A* u), where A* is equivalent to A, 
  ;; u is a member of (cdr A*) which contains x, and
  ;; if (car A) is &  then u has the form (or P C) or (-> P C), and
  ;; if (car A) is OR then u has the form (& P C),
  ;; where x occurs in C but NOT in P.

  ;;%%%% We are now just calling DISJUNCTIVE-VE-UNIT, etc from here.
  ;;     We might want to combine these into one function later.

  (cond ((eq '&  (car A)) (disjunctive-ve-unit A x))
	((eq 'or (car A)) (conjunctive-ve-unit A x))
	(t nil)))

(defun disjunctive-unit (L x)    ;called by VE
  ;; Sellect a member of L which contains x and is of the form (or a ...)
  ;; or  (-> a b)

 (cond ((null L) nil)
       ((not (occur x L)) nil)
       ((member (caar L) '(or ->)) (car L))
       (t (disjunctive-unit (cdr L) x)) ))

(defun disjunctive-ve-unit (A x)   ;Called by disjunct-conjunct-unit 26Feb90
  (let (u u*)
    (cond ((some #'(lambda (u1) 
        (and (eq-car u1 'or) 
             (some #'(lambda (u2) (and (not (occur x u2)) (setq u u1)
	                               (setq u* (list 'or u2 
					   (remove u2 u :test 'equal)))))
		   (cdr u1))))
		 (cdr A))
           ;;For A = (& P (or Ax Bc Cx) Q), we get u  = (or Ax Bc Cx), 
           ;;                                      u* = (or Bc (or Ax Cx))
           ;;And ( (or Ax Bc Cx) (or Bc (or Ax Cx)) )  is returned.
	   (list u u*))

          ((some #'(lambda (u1) (and (eq-car u1 '->)
				     (or (and (not (occur x (cadr u1)))
					      (setq u u1) (setq u* u))
					 (and (not (occur x (caddr u1)))
					      (setq u u1)
					      (setq u* (list '-> (not-on (caddr u))))))))
		 (cdr A))
           ;;For A = (& P (-> (& Ax Cx) Bc) Q), we get u = (-> (& Ax Cx) Bc),
           ;;                                     u*  = (-> ~Bc (or ~Ax ~Cx))
           ;;And ( (-> (& Ax Cx) Bc) (-> ~Bc (or ~Ax ~Cx)) ) is returned.
	   (list u u*))
          
	  (nil nil) ;Other cases later. 
	  )))

(defun conjunctive-ve-unit (A x)   ;Called by disjunct-conjunct-unit 26Feb90
  (let (u u1*)                     ;Similar to disjunctive-ve-unit
    (cond ((some #'(lambda (u1) 
        (and (eq-car u1 '&) 
             (some #'(lambda (u2) (and (not (occur x u2)) (setq u u1)
	                               (setq u* (list '& u2 
					   (remove u2 u :test 'equal)))))
		   (cdr u1))))
		 (cdr A))
           ;;For A = (or P (& Ax Bc Cx) Q), we get u  = (& Ax Bc Cx), 
           ;;                                      u* = (& Bc (& Ax Cx))
           ;;And ( (& Ax Bc Cx) (& Bc (& Ax Cx)) ) is returned.
	   (list u u*))

	  (nil nil) ;Other cases later. 
	  )))

(defun eligible (x A &optional S)	;used in ELIGIBLES
					;is x eligible in A?
  (cond ((not (occur x A)) t)
	((eq x A) S) ;an eligible variable has to be in the scope of < or <=
        ((member (car a) '(< <=)) 
         (eligible x (cond ((cddr a) (minus-on-reduce (cadr A) (caddr A))) 
                           (t (cadr A))) t))
	((member (car a) '(& or not -> + -))
	 (every #'(lambda (u) (eligible x u S)) (cdr A)) )) )
 
(defun eligibles (A &optional L)	
  ;;used in INEQUALITY-PROVE, OPTIONS, ELIGIBLES-ALL
  ;;the list of eligibles variables in A, and occuring in L (unless L is nil)

  (remove-if-not #'(lambda (x) (eligible x A))   
		 (cond (L L) (T (free-variables-in A)))))

(defun eligibles-all (H C YET &optional L)  ;Called by OPTIONS
  ;;This gives the list of variables which are eligible in C or H and not
  ;; shielded in YET. (And which are members of L, if L is not NIL).
  (let (E SY)
       (cond ((null (setq E (union-w (eligibles C L) (eligibles H L)))) nil)
             ((null (setq SY (shieldeds-yet YET E))) E)
             (t (set-difference E SY)))))

(defun eligible-split (A x &optional (parity t)) ;Called by VE-something
  ;;This splits the formula a into a disjunct (AA v AA') where x is 
  ;; eligible in AA but not in AA'. See example at end of the function.

  (mapcar #'reduce-el-split (eligible-split2 A x parity)))

#|For example, if A is the formula,

(-> (< (- (E)))
    (& (< (- D))
       (-> (< (+ (|| (+ (XS D) (- (A)))) (- D)))
           (& (< (- E2)) (<= (+ (- (D1S1 E2)) D))
              (<= (+ (- (/ (E) 2)) E2)) (< (- E1))
              (<= (+ (- (D1S E1)) D)) (<= (+ (- (/ (E) 2)) E1))))))
      
and if we call (ELIGIBLE-SPLIT A 'D t), and simplify the result 
using the function REDUCE-EL-SPLIT, we get the following  pair (AA AA'),

((-> (< (- (E)))
     (& (< (- D))
        (& (< (- E2)) (<= (+ (- (D1S1 E2)) D))
           (<= (+ (- (/ (E) 2)) E2)) (< (- E1)) (<= (+ (- (D1S E1)) D))
           (<= (+ (- (/ (E) 2)) E1)))))
 (& (< (- D)) (<= (+ (- (|| (+ (XS D) (- (A))))) D))))

Note that (1) D is elligible in AA, (2) but not in AA', (3) the disjunction
(or AA AA') is equivalent to A, and (4) AA is the "largest such D-elligible 
piece" that can be taken from A.

Furthermore, if we VE (variable-eliminate) D from AA to get AA*, then
the disjunction (or AA* AA') is also equivalent to A, and is simplier
in that it has one fewer variables (namely D).
|#

(defun eligible-split2 (A x parity)
  (let (A* A1 A2 H* C*)
    (cond
      ((eligible x A) (cond (parity (list A nil)) (t (list A t))))
      ((atom A) (cond (parity (list A nil)) (t (list A t))))
         ;;Probably covered by the previous line.
      ((eq '& (car A)) 
       (setq A* (mapcar #'(lambda (u) (eligible-split2 u x parity)) (cdr A)))
       (cond ((and parity (some #' (lambda (u) (null (car u))) A*)) 
              (list nil A))
             ((and (not parity) (every #' (lambda (u) (null (car u))) A*)) 
              (list t A))
             (t (setq A1 (cond (parity (cons '&  (mapcar 'car A*)))
                               (t      (cons 'or (mapcar 'car A*)))))
                (setq A2 (cond (parity (remove A1 (or-and-together A*)
                                                  :test 'equal))
                               (t      (cons 'or (mapcar 'cadr A*))))) 
                (list A1 A2) )))

      ((eq 'or (car A)) 
       (setq A* (mapcar #'(lambda (u) (eligible-split2 u x parity)) (cdr A)))
       (cond ((and (not parity) (some #' (lambda (u) (null (car u))) A*)) 
              (list t A))
             ((and parity (every #' (lambda (u) (null (car u))) A*)) 
              (list nil A))
             (t (setq A1 (cond (parity (cons 'or (mapcar 'car A*)))
                               (t      (cons '&  (mapcar 'car A*)))))
                (setq A2 (cond (parity (cons 'or (mapcar 'cadr A*)))
                               (t   (remove A1 (and-or-together A*)
                                               :test 'equal))))
                (list A1 A2) )))

      ((eq '-> (car A))
       (setq H* (eligible-split2 (cadr A) x (not parity)))
       (setq C* (eligible-split2 (caddr A) x  parity))
       (cond ((and parity (null (car H*)) (null (car C*))) (list nil A))
	     ((and (not parity) (null (car H*)) (null (car C*))) (list t A))
	     (t (setq A1 (list '-> (car H*) (car C*)))
		(setq A2 (cond (parity (list '-> (cadr H*) (cadr C*)))
			       (t (and-on (list '-> (car H*) (cadr C*))
					  (and-on (list '-> (cadr H*) (car C*))
						  (list '-> (cadr H*) 
							(cadr C*)))))))
		(list A1 A2) )))

      ((eq 'not (car A))
       (setq A* (eligible-split2 (cadr A) x (not parity)))
       (cond ((and parity (null (car A*))) (list t A))
	     ((and (not parity) (null (car A*))) (list nil A))
	     (t (list (not-on (car A*)) (not-on (cadr A*))) ) ))

      (parity (list nil A))
      (t      (list t   A)) )))
#|  this was disabled on 4/4/89 so that it would load.  fix it later
(defun E-Eligible (H C Yet)  ;Called by OPTIONS
  ;;This means "Eligible-Eligible".  It discovers whether C or H has
  ;; a variable x which is eligible to be eligible, in the following sense-
  ;; (ELIGIBLE-SPLIT C x) returns (A A') where A is not NIL,
  ;; and prints x and (A A'), and whether x is in C or H.
  ;; It also checks that x is not shielded in YET. And returns (x A A').
  ;;(3/18/88) This should be extended to verify that x is also eligible in YET.

  (let (VC VH V W)
    (setq W nil)

    (cond ((and (setq VC (free-variables-in C)) 
                (null (setq V (set-difference VC (shieldeds-yet yet VC)))))
	   (print '(C has no free variables not shielded in YET)) nil)
          
          ((some #'(lambda (x) (and (setq W (eligible-split C x t)) 
                                    (occur x (car W)) (setq W (cons x W)))) V)
	   (format t "~% C can be written as a disjunction of the two terms
~% ~s ~%and~% ~s ~%~% where the variable ~s is eligible in the first.~% ~%" 
(cadr W) (caddr W) (car W)) (return-from E-eligible W))

          (t (print '(C has no E-Eligible variables)) nil))

    (cond ((and (setq VH (free-variables-in H)) 
                (null (setq V (set-difference VH (shieldeds-yet yet VH))))
           (print '(H has no free variables not shielded in YET)) nil))

          ((some #'(lambda (x) (and (setq W (eligible-split H x nil))
                                    (occur x (car W)) (setq W (cons x W)))) V)
           (format t "~% H can be written as a disjunction of the two terms
~% ~s ~%and~% ~s ~%~% where the variable ~s is eligible in the first.~% ~%" 
(cadr W) (caddr W) (car W)) W)
          (t (print '(C has no E-Eligible variables)) nil)) ))
|#

(defun Eliminate-variable () ;Called by OPTIONS
  ;;7/7/88 Change to account for changed form of H, ie, (use h-gu H= Hv).

  ;;Eliminate a variable x in C or H, as directed by the user.
  ;;The default case is the first free variable that is eligible in
  ;; a disjunct of C.
 (declare (special h c tl anc lt yet))
 (let (x W A A* AA C*)

  (format t "~%      ELIMINATE A VARIABLE ~%")
  (format t "~% Do you wish to designate the variable?  
  If so type the variable, if not type the word  NO ~%")
  (setq x (read-h))
  (setq x (cond ((eq 'no x) nil) (t x)))
  (cond (x (cond ((and (setq W (eligible-split C x)) (occur x (car W)))
                  (setq W (cons x W)))
                 (t (format t "~% ~s is not E-Eligible in C~%" x)
		    (if (y-or-n-p "Do you wish to give another variable? Type-")
			(return-from Eliminate-variable (Eliminate-variable))
                      (return-from Eliminate-variable (options))))))
        (t (cond ((setq W (E-Eligible H C YET)))
                 (t (print '(No E-Eligible variable in C))
                    (return-from Eliminate-variable (options))))))

  (setq x (car W)) (setq A (reduc* (cadr W))) (setq A* (caddr W))

;;;NOTE:  be concerned about PARITY in the call to (VE x A)

  (setq AA (VE x A))
  (setq C* (or-on AA A*))
  (format t "~%The variable ~s has been eliminated by VE from ~% 
     ~s ~% to obtain ~% ~s ~%" x C C*)
  (format t "_______________________________________________________")
  (setq C C*) ))      ;we must still do H
    
(defun flip-h (v &optional sh)  ;called by inequality-prove  3 June 89
  ;;Here v is a set of variables in hv (= (fourth h)). 
  ;;This routine selects a variable x from v, finds a conjuct cj of hv 
  ;; which contains x, and moves its negation to "C", the goal position;  
  ;; and add ~C to H.  Then proceeds with the proof.
  ;;If the rest of the proof is successful, OK, else another x is selected
  ;; from v, etc.
  ;;There are two reasons for moving a conjunct cj of hv containing x to C:
  ;; (1) The variable x is eligible in the whole formula, and cj is moved
  ;;     to C so that x can be eliminated by VE.
  ;; (2) A term of cj shields x, and cj is being moved to C so that that 
  ;;     term can be removed by STR (using REPLACE-A). 
  ;;     In this case the parameter SH will be T.

  (declare (special h c tl anc lt yet))
  (let ((hv (fourth h)) x cj h*)    

    (cond ((null v) (return-from flip-h nil)))
           ;;none of the variables of v have succeeded.
    (setq x (car v))
    (cond ((and (listp hv) (eq '& (car hv))) 
	   (some #'(lambda (u) 
		     (cond ((and sh (member x (shieldeds u))) (setq cj u))
			   ((occur x u) (setq cj u)) 
			   (t nil)))
		 hv)
	   (setq h* (clear-& (remove cj hv :test 'equal))))
	  (t (setq cj hv) (setq h* nil)))

    (setq h (subst h* hv h :test 'equal))
    (setq h (and-on-h h (push-nots-in-ve (not-on C))))
    (setq c (push-nots-in-ve (not-on cj))))

    (cond ((inequality-prove h c (append tl '(flip-h)) anc lt yet))
	  ((imply h c (append tl '(flip-h)) anc lt yet))
	  (t (flip-h (cdr v) sh))))

    
(defun flip-yet (v &optional sh)  ;called by inequality-prove  3 June 1989
                                  ;&optional sh  added 29 July 89
  ;;Here v is a set of variables in yet.
  ;;This routine selects a variable x from v, finds a unit u1 of yet 
  ;; which contains x, and moves that unit (with H C tl etc) to the
  ;; prover part, and moves the current H C tl etc to u1's position 
  ;; in yet.  Then proceeds with the proof.
  ;;If the rest of the proof is successful, then OK, 
  ;; else another x is selected from v, etc.
  ;;There are two reasons for moving a unit u1 containing x from yet:
  ;; (1) The variable x is eligible in u1, and u1 is moved
  ;;     to C so that x can be eliminated by VE.
  ;; (2) A term of u1 shields x, and u1 is being moved to C so that that
  ;;     term can be removed by STR (using REPLACE-A). 
  ;;     In this case the parameter SH will be T.

  (declare (special h c tl anc lt yet))
  (let (x u1 u2 h* c* tl* anc* lt* yet* h** c** u)

    (cond ((null v) (return-from flip-yet nil)))
           ;;none of the variables of v have succeeded.
    (setq x (car v))

    (cond ((some #'(lambda (u) (setq u1 (and-on (second u) (third u)))
		     (cond ((and sh (member x (shieldeds u))) (setq u1 u))
			   ((occur x u) (setq u1 u)) 
			   (t nil)))
		 yet))
	  (t (return-from flip-yet nil)))     ;alarm

    (cond ((occur 'sub-here (second u1))
	   (setq h** (list 'sub-here h)) (setq c** c))
	  ((occur 'sub-here (third u1))
	   (setq h** h) (setq c** (list 'sub-here c)))
	  (t (setq h** h) (setq c** c)))
    (setq u2 (list T h** c** tl anc lt 'flip-yet))
    (setq yet* (subst u2 u1 yet :test 'equal))
    (setq u (apply-substitution-restricted (car u1) (cdr u1)))
    (multiple-value-setq (h* c* tl* anc* lt*) u)
    (cond ((inequality-prove h* c* tl* anc* lt* yet*))
	  ((imply h* c* tl* anc* lt* yet*))
	  (t (flip-yet (cdr v) sh)))))
#|
(defun flip-h-the-variable (x)  ;called by inequality-prove  30 May 89
  ;;Here we find the conjuncts of H that contains x, and move their negation to
  ;; "C", the goal position;  and add ~C to H.  Then proceed with the proof.
  ;;There are three places in INEQUALITY-PROVE that call the function:
  ;; (1) where the variable x is eligible in C but shielded in H. In this case
  ;;     the shielding conjuncts are moved to C so that they can be removed.
  ;; (2) where the variable x is eligible in H and C has no eligible variable.
  ;;     In this case (the negation of) the  conjuncts of H containing x are 
  ;;     flipped over to C and ~C to H as before.  
  ;; (3) where the variable x occurs in H, and C has no variable. 
  ;;     This is similar to (2).

  (declare (special h c tl anc lt yet))
  (let ((hv (fourth h)) h*)
    (cond ((atom hv) alarm)
          ((eq '& (car hv))
           (setq h* (remove-if-not #'(lambda (u) (occur x u)) (cdr hv)))
	   ;;h* consists of those members of hv containing x 
           (setq h** (set-difference (cdr hv) h*))
           (setq h*  (cond ((cdr h*)  (cons '& h*))  (t (car h*))))           
           (setq h** (cond ((cdr h**) (cons '& h**)) (t (car h**))))
	   (setq h (subst h** hv h :test 'equal))
           (setq h (and-on-h h (not-on C)))
           (setq c (not-on h*)))
          (t (setq h (subst nil hv h :test 'equal))
             (setq h (and-on-h h (not-on C)))
             (setq c (not-on hv))))
    (cond ((inequality-prove h c (append tl '(flip-h)) anc lt yet))
	  (t (imply h c (append tl '(flip-h)) anc lt yet)))))
|#

(defun ground-prover (A &optional ineq-list)    ;Called by Inequality-Prove
  ;; A is an inequality theorem. It may or may not be ground, but it 
  ;; processed as if it were ground.
  ;; INEQ-LIST, if non-empty,  is an Unsat set of ground ineq literals
  ;; This version uses Don Simon's Matrix inequality prover,
  ;;  somewhat like the sup-inf method,
  ;;  implementation in dons-ground-Prover.lisp
  ;;
  ;; First, we convert from unary < and <= forms to binary, if needed.
  ;;  Eg, (< (+ a (-b))) goes to (< (+ a (-b)) 0).
;;%%remove the or t part.  Later 29 July 89, I do not understand this remark.
    (cond ((or (not *ground-prover-light*)
	       (and (null A) (null ineq-list)))
	   (return-from ground-prover nil)))

    (setq A (unary-to-binary< A)) 
    (setq ineq-list (mapcar #'unary-to-binary< ineq-list))
    (setq A (minus-to*-1 A))
    (setq ineq-list (mapcar #'minus-to*-1 ineq-list))

    (setq ineq-list (cond (ineq-list (list ineq-list))
		     (t (prepare-for-sup-inf A))))
    ;; Now, each list in INEQ-LIST is an Unsat set of ground ineq literals
    (every #'(lambda (literals) (ground-p literals)) ineq-list))

(defun ground-p (literals)    ;called by ground-prover
  (let (ans E lits)
    (cond ((multiple-value-setq (ans E) (find-contradiction literals)) ans)
	  (E (setq lits (g-sublis E (set-difference literals E :test 'equal)))
             (ground-p lits))    ;substitute equals and try again
	  (t nil))))

(defun g-sublis (E lits)  ;called by ground-p
  ;;E is a list of the form (<= a b) (<= b c) etc
  ;; where a=b, b=c, etc.  Replace a by b, and b by c, etc in lits
  (let (u)
    (cond ((null E) lits)
	  ((occur (cadr (setq u (car E))) (caddr u))  ;eg, u = (<= x (f x))
	   (g-sublis (cons (cons '<= (reverse (cdr u))) (cdr E)) lits))
	  (t (g-sublis (cdr E) (g-subst (caddr u) (cadr u) lits ))))))

(defun g-subst (y x lits)  ;called by g-sublis
  (cond ((equal lits (setq lits (subst y x lits :test 'equal))) lits)
	(t (g-subst y x lits))))

(defun prepare-for-sup-inf (A)   ;called by GROUND-PROVER
  ;;This converts the negation of the fomula A to DNF.
  ;;A is assumed to be a GROUND formula in FOL.

    (setq A (convert-implication-signs (list 'not A)))
    (setq A (push-nots-in-ve A))
    (setq A (DNF A))     
    (setq A (remove-or&-&-signs A)))

(defun remove-or&-&-signs (A) ;called by PREPARE-FOR-SUP-INF
  ;;A is assumed to be in CNF, the output is a set of clauses
  (cond ((atom A) (list (list A)))
        ((eq 'or (car A))
         (mapcan #'(lambda (u) (remove-or&-&-signs u)) (cdr A)))
        ((eq '& (car A)) (list (cdr A)))
        (t (list (list A)))))

(defun convert-implication-signs (A) 
  ;;Ie, (-> P Q) goes to (or (not P) Q)
  (cond ((atom A) A)
        ((eq '-> (car A)) 
         (list 'or (list 'not (convert-implication-signs (cadr A)))
                   (convert-implication-signs (caddr A))))
        ((eq 'not (car A)) 
	 (cond ((atom (cadr A)) A)
	       ((eq '-> (caadr A))  ; ~(H->C) goes to (H & ~C)
		(setq A (cadr A))
		(setq A (list '& (cadr A) (list 'not (caddr A))))
		(convert-implication-signs A))
	       (t (list 'not (convert-implication-signs (cadr A))))))

        ((member (car A) '(& or))
         (cons (car A) (mapcar #'convert-implication-signs (cdr A))))

        (t A)))

(defun push-nots-in-ve (A)   ;called by PREPARE-FOR-SUP-INF, and by VE
  ;;A has no -> signs, if called by PREPARE-FOR-SUP-INF.
  ;;eg, (not (& A (or B C))) goes to  (or (not A) (& (not B) (not C)))
  ;;    (not (< a b)) goes to (<= b a) etc
  (cond ((atom A) A)
        ((eq 'not (car A))
         (cond ((atom (cadr A)) A)
               ((eq 'not (caadr A))       ;(not (not A)) goes to A
		(push-nots-in-ve (cadadr A)))  

               ((eq '& (caadr A))
                (cons 'or (mapcar #'(lambda (x) 
                            (push-nots-in-ve (list 'not x))) (cdadr A))))

               ((eq 'or (caadr A))
                (cons '& (mapcar #'(lambda (x) 
                            (push-nots-in-ve (list 'not x))) (cdadr A))))

	       ((eq '-> (caadr A))
		(list '& (push-nots-in-ve (cadadr A))
		         (push-nots-in-ve (list 'not (caddr (cadr A))))))

	       ((eq '<= (caadr A)) 
		(cond ((cddadr A) (list '<  (caddr (cadr A)) (cadadr A)))
		      (t          (list '<  (reduc* (list '- (cadadr A)))))))
	        ;;Note that < and <= may be either BINARY or UNARY functions.
	       ((eq '<  (caadr A)) 
		(cond ((cddadr A) (list '<= (caddr (cadr A)) (cadadr A)))
		      (t          (list '<= (reduc* (list '- (cadadr A)))))))

               (t A)))
        ((member (car A) '(& or)) 
         (cons (car A) (mapcar #'push-nots-in-ve (cdr A))))

        (t A)) )

(defun dnf (A)      ;put A in Disjunctive Normal Form.
  ;;Assume that A has no -> signs, and has nots pushed in.
  ;;Ie, (or (& A B) C)  goes to  (& (or A C) (or B C))
  (let (AL)
    (cond ((atom A) A)
          ((and (member (car A) '(& or)) 
                (setq A (cons (car A) (mapcar #'dnf (cdr A)))) nil))
          ((eq 'or (car A)) 
           (setq AL (dnf-or (cdr A)))      ;remove the other or's
           (cons 'or AL))

          ((eq '& (car A)) 
           (cond ((occur 'or (cdr A))      ;permute or's and &'s and recurse
                  (dnf-&-or A))
                 (t (setq AL (dnf-& (cdr A)))     ;remove the other &'s
                    (cons '& AL))))
          (t A))))

(defun dnf-or (AL)   ;called by dnf
  ;;remove the top level or's (flatten)
    (cond ((null AL) nil)
          ((atom (car AL)) (cons (car AL) (dnf-or (cdr AL))))
          ((eq 'or (caar AL))
           (dnf-or (append (cdar AL) (cdr AL))))
          (t (cons (car AL) (dnf-or (cdr AL))))))

(defun dnf-& (AL)   ;called by dnf
  ;;remove the top level &'s (flatten)
    (cond ((null AL) nil)
          ((atom (car AL)) (cons (car AL) (dnf-& (cdr AL))))
          ((eq '& (caar AL))
           (dnf-& (append (cdar AL) (cdr AL))))
          (t (cons (car AL) (dnf-& (cdr AL))))))
          
(defun dnf-&-or (A &optional AF)   ;called by dnf
  ;;A starts with the form (& A B ...)
  (cond ((null A) AF)
        ((atom (car A)) (dnf-&-or (cdr A) (append AF (list (car A)))))
        ((eq 'or (caar A))
         (setq A (mapcar #'(lambda (x) (append AF (cons x (cdr A)))) (cdar A)))
         (setq A (cons 'or A))
         (dnf A))
        (t (dnf-&-or (cdr A) (append AF (list (car A)))))))


(defun l-le (S1 S2)         ;Called by VE-make
  (if (and (eq '<= S1) (eq '<= S2)) '<= '<))
  
(defun l-le* (S1 S2)         ;Called by VE-make
  (if (or (eq '<= S1) (eq '<= S2)) '<= '<))
  
 
(defun left-right-sides (x c)		;called by VE-make
  ;; This returns the list W = (LS RS NS), where LS is the list of units
  ;; of the form (< ls) where for u in C, u can be put in the form (ls < x).
  ;; Similarly, RS is the list of units of the form (< rs) where for some u 
  ;; in C, u can be put in the form (x < -rs) (Note the -).
  ;;  NS are the remaining units of C.
  ;; For example if C = ( [< (-x)] [<= x] [<= (+ a b (-x) c)] [< (+ c x d)]
  ;;                      [< (+ e f)] [<= a]),
  ;; then  W = (LS RS NS) is returned with 
  ;;            LS = ( [< 0] [<= (+ a b c)]),
  ;;            RS = ( [<= 0] [< (+ c d)]),  and
  ;;            NS = ( [< (+ e f)] [<= a])
  
  (let (mx LS RS NS)
    (setq mx (list '- x))
    (mapc
      #'(lambda (u) 
	  (cond 
	   ((equal mx (cadr u))		;case where u = (< (-x)), means -x<0
	    (setq LS (cons (list (car u) 0) LS)))
	   ((eq x (cadr u))		;case where u = (< x) ,  means x<0
	    (setq RS (cons (list (car u) 0) RS)))
           ((atom (cadr u)) (setq NS (cons u NS)))
	   ((member mx (cdadr u) :test 'equal) ;case where u = (< (+ a (-x) b))
	                                       ;        means  (+ a -x b)< 0
	    (setq LS 
                  (cons (list (car u) (remove mx (cadr u) :test 'equal)) LS)))
	   ((member x (cdadr u))	;               (+ a x b)< 0
	    (setq RS (cons (list (car u) (remove x (cadr u))) RS)))
	   (t (setq NS (cons u NS))) )) c)
    (list LS RS NS) ))

(defun minus-it (a)   ;called by left-right-sides
 (reduc (list '- a))) 

(defun minus-to*-1 (A)    ;called by GROUND-PROVER
  ;;convert expression of the form (- A) to (s* -1 A) for use in
  ;; FIND-CONTRADICTION
  (cond ((atom A) A)
	((eq '- (car A)) (list 's* -1 (cadr A)))
	(t (cons (minus-to*-1 (car A)) (minus-to*-1 (cdr A))))))

(defun not-formula-< (A)  ;Called by Prepare-for-sup-inf; no not used
  ;;Eg, (not B) goes to B, B goes to (not B), (< a b) to (<= b a), etc.
  (cond ((atom A) (list 'not A))
        ((eq 'not (car A)) (cadr A))
	((eq '<= (car A)) (list '<  (caddr A) (cadr A)))
	((eq '<  (car A)) (list '<= (caddr A) (cadr A)))
        (t (list 'not A))))

(defvar ex-ve nil)
(setq ex-ve '( (< (- x)) (<= x) (<= (+ a b (- x) c)) (< (+ c x d))
                     (< (+ e f)) (<= a)))
(defun orify (l)			;used in VE
  ;;convert a list into a conjunction
  (cond ((null (cdr l)) (car l))
        (t (cons 'or l))))
 
(defun plus-on (a b) 
  (setq a (clear-+ a)) (setq b (clear-+ b))
  (cond ((and (numberp a) (zerop a)) b)
	((and (numberp b) (zerop b)) a)
	(t (reduc (cond ((and (not (atom a)) (eq '+ (car a)))
			 (cond ((and (not (atom b)) (eq '+ (car b)))
				(append a (cdr b)))
			       (t (append a (list b))) ))

			((and (not (atom b)) (eq '+ (car b)))
			 (cons '+ (cons a (cdr b))))

			(t (list '+ a b)) )))))

(defun prepare-u (u x)   ;called by VE
  ;; u is either (or P A ...)  or (-> P A)
  ;; Rearrange u so that u = (or P A) or u = (-> P A) and x does not occur 
  ;;  in P, and return W = (P A).  If not possible, return nil.
  (let (P A)
   (cond

    ((eq '-> (car u)) 
     (cond
       ((not (occur x (cadr u))) (cdr u))
       ((not (occur x (caddr u))) (list (not-on (caddr u)) (not-on (cadr u))))
       (t nil) ))                     ; if x occurs in both, return nil

    ((eq 'or (car u)) 
     (cond
       ((not (occur x (cadr u))) (list (cadr u) (orify (cddr u))))
       ((setq P (some #'(lambda (uu) (and (not (occur x uu)) uu)) (cddr u)))
        (setq A (orify (remove P (cdr u) :test 'equal)))
        (list P A)) )) )))

(defun remove-shield (c* vc &optional reject-list)  ;called by inequality-prove
  ;;Remove a shielding term from c* (with variables in vc) using REPLACE-A, 
  ;; and prove the resulting formula.   30 May 89
  (declare (special h c tl anc lt yet))
  (let (w tm h* theta yet* tl* lambda)
    (cond ((null vc) nil)
          ((not (setq tm (a-shielding-term c* vc reject-list))) nil)
          ((not (cadr (setq w (replace-a (and-on (second h) (fourth h))
					 c* tm reject-list))))
	   (remove-shield c* vc (cons tm reject-list)))
	  (t (setq h* (apply-substitution (cadr w) h))
	     (setq c* (car w))
	     (setq theta (cadr W))
             (setq yet* (mapcar #'(lambda (u) 
                (cons (compose (car u) theta) (cdr u))) yet))
             ;;compose theta with existing subs in yet.  
	     ;;(Should we worry about variable conflicts, as in prove-the-pre?)
	     (setq tl* (append tl (list 'rs)))
	     (cond ((setq lambda (inequality-prove h* c* tl* anc lt yet*))
		    (compose theta lambda))
		   ((setq lambda (imply h* c* tl* anc lt yet*))
		    (compose theta lambda))
		   (t nil))))))

(defun remove-sub-here (A)  ;called by VE-the-variable
  (cond ((atom A) A)
	((eq 'sub-here (car A)) (cadr A))
	(t (cons (remove-sub-here (car A)) (remove-sub-here (cdr A))))))

(DEFUN SHIELDEDS (A &OPTIONAL (L NIL))	;USED IN INEQUALITY-PROVE
  ;;the list of variables shielded in A, and occuring in L (unless L is nil)

  (cond ((atom A) nil)
        ((member (car A) '(< <= & or not -> + - ))
         (mapcan #'(lambda (u) (shieldeds u L)) (cdr A)))
        (t (let ((V (free-variables-in A)))
                (cond (L (set-intersection L V)) (t V)))) ))

(defun shielding-terms (x A &optional print L) ;called by options
  ;;Prints the list of all shielding terms of x in A.
  (let (sh)
    (cond ((setq sh (a-shielding-term x A L))       ;should x be (list x)?
	   (shielding-terms x A print (cons sh L)))
	  (print 
	   (format t "~% The Shielding terms of ~s in C are: ~2% ~s~2%" x L)L)
	  (t L) )))

(defun a-shielding-term (c* vc &optional L)  ;Used by SHIELDING-TERMS
  ;;find a shielding term tm in c*, that shields a variable in vc, for which 
  ;; tm is not in L
       (cond ((atom c*) nil)
	     ((member (car c*) '(< <= & or not -> + -))
	      (some #'(lambda (u) (a-shielding-term u vc L)) (cdr c*)))
             ((and (not (occur c* L)) 
		   (some #'(lambda (x) (occur x c*)) vc)) c*)
	     (t nil) ))

(defun shieldeds-yet (A &optional (L nil)) ;used in INEQUALITY-PROVE
  ;;this returns the list of variables shielded in the H&C slots 
  ;;of the units of A (= YET), and occuring in L (unless L is nil)
  ;;Recall that YET is a list of units u of the form 
  ;; (theta h c tl anc lt comment)

  (setq A (remove-sub-here A))
  (crunch (mapcan #'(lambda (u) (shieldeds (and-on (cadadr u) (caddr u)) L))
                  A)))

(defun unary-to-binary< (A)  ;called by GROUND-PROVER
  (cond ((atom A) A)
	((member (car A) '(< <=))
	 (cond ((cddr A) A) (t (list (car A) (cadr A) 0))))
	(t (cons (unary-to-binary< (car A))
		 (unary-to-binary< (cdr A))))))

(defun VE-s (ec A)                 ;Called by INEQ-PR
  ;;This just repeatedly calls VE until all variables of ec are eliminated.
  (let (A*)
    (setq A* A)
    
    (mapc #'(lambda (x) (setq A* (VE x A* t))) ec)
    A*))

(defun VE (x A &optional (parity t))   ;automatic parity = t, added 26Feb90
  ;;Called by VE-the-variable and Remove-variable (See Remove.lisp, 12/4/87)
  ;; Elliminate the variable x from A (x is assummed to be Eligible in A)
  ;; by the process in the CADE 1980 paper, and by using the VE-Reduction
  ;; Theorem (8/25/87):
  ;;
  ;;    (SOME x [(P -> A(x)) & Q(x)])   <-->
  ;;
  ;;         (SOME x Q(x))  &  (SOME x (P -> [A(x) & Q(x)]))

 (let (u Q W P C A*)
  (cond
   ((atom A) A)
   ((not (occur x A)) A)
   ((member (car A) '(< <=)) (cond (parity t) (t '(FALSE))))
   ((eq 'not (car A)) (cons 'not (VE x (cadr A) (not parity))))
   ((or (and (eq 'or (car A)) parity) (and (eq '& (car A)) (not parity)))
    (setq A* (cons (car A) 
		    (mapcar #'(lambda (u) (VE x u parity)) (cdr A))))
    (cond ((member t A*) t) 
	  ((and (setq A* (cond (parity (clear-or (remove nil A*)))
			       (t      (clear-&  (remove nil A*))))) nil))
	  (t A*)))

   ((and (eq '-> (car A)) parity)
    (list '-> (VE x (cadr A) (not parity)) (VE x (cadr A) parity)))

   ((or (and (eq '& (car A)) parity) (and (eq 'or (car A)) (not parity)))
    (cond ((every #'<p (cdr A)) (VE-make x A)) ;A is ready for VE
	  ((setq W (constant-part< (cdr A) x))   ;26Feb90
	   ;;W has the form NIL or (Ac Ax) where (cdr A) = (Ac U Ax) 
           ;;  and x does not occur in Ac and each member of Ax is an ineq unit
	   (setq A* (VE-make x (cons (car A) (cadr W))))
	   (setq W (cons (car A) (cadr W)))
	   (cond ((eq '& (car A)) (and-on W A*))
		 (t               (or-on W A*))))

	  ((setq W (disjunct-conjunct-unit A x))
	   ;;W has the form (u u*), where u is a member of (cdr A) containing x
	   ;;  and u* is equivalent to u. 
	   ;;If parity=t (ie, A has the form (& ...)), then u* has the form
	   ;;  (or P C) or (-> P C), where x occurs in C but NOT in P.
	   ;;Else u* has the form  (& P C).

	   (setq u (car W)) (setq W (cadr W))
	   (setq Q (remove u A :test 'equal))
	   (setq Q (cond (parity (clear-& Q)) (t (clear-or Q))))
	   (setq P (cadr W)) (setq C (caddr W))
	   (setq A* (and-on (list (car u) P (VE x (and-on C Q) parity)) 
			    (VE x Q parity)))
	   ;;This applies the VE-Reduction Theorem, to get
	   ;;    (P -> (VE x [A(x) & Q(x)]))  &  (VE x [Q(x)])
	   (reduc* A*) )
	  ((equal A (setq A* (cond (parity (dnf A)) (t (cnf A))))) nil)
	  (t (VE x A* parity))))
   ((and (eq '-> (car A)) (not parity))
    (cond ((not (occur x (cadr A))) 
	   (list '-> (cadr A) (VE x (caddr A) parity)))
	  ((not (occur x (caddr A)))
	   (list '-> (VE x (cadr A) (not parity)) (caddr A)))
	  ((and (or (<p (cadr A)) 
		    (and (&p (cadr A)) (every #'<p (cdadr A))))
		(or (<p (caddr A)) 
		    (and (orp (caddr A)) (every #'<p (cdaddr A)))))
           ;;Example:[(x<a & x<c -> x<=b) --> C] goes to [(a<=b or c<=b) --> C]
	   ;; because (x<a & x<c -> x<=b)  goes to  (a<=b or c<=b), parity nil
	   (VE x (push-nots-in-ve 
		  (and-on (not-on (cadr A)) (caddr A))) parity))
	  
	  ((&p (caddr A))                          ;eg, (x<a -> x<=b & x<=d)
	   (set A* (cons '& (mapcar #'(lambda (u) (list '-> (cadr A) u)) 
				    (cdaddr A))))   ;to (x<a -> x<=b) & 
	                                            ;   (x<a -> x<=d)
	   (VE x A* parity))
	  ((orp (cadr A))                          ;eg, (x<a or x<c -> x<=b)
	   (set A* (cons '& (mapcar #'(lambda (u) (list '-> u (caddr A))) 
				     (cdadr A)))))  ;to (x<a -> x<=b) & 
	                                            ;   (x<c -> x<=b)
	  ((equal A (setq A* (dnf A))) nil)
	  (t (VE x A* parity))))
   (t nil) )))

(defun <p  (A) (and (listp A) (member (car A) '(< <=))))
(defun &p  (A) (and (listp A) (eq '&  (car A))))
(defun orp (A) (and (listp A) (eq 'or (car A))))
(defun notp (A) (and (listp A) (eq 'not (car A))))

(defun VE-make (x A)    ;Called by VE
  ;; This makes up the actual results of (VE x A)
  ;; A is either a conjunctive form with inequality literals,
  ;;          or a disjunctive form with inequality literals.
  ;; This expects unary < and <=
  ;; For example, (or (< x (a)) (< (b) x))  occurs as
  ;;              (or (< (+ (- x) (a))) (< (+ (- (b)) x))) 
  ;; and  (< (+ (- (b)) (a)))  is returned (equivalent to (< (b) (a)))
  ;;  goes to  (< (b) (a))
 (let (W LS RS NS A*)

  (setq W (left-right-sides x (cond ((member (car A) '(& or)) (cdr A))
				    (t A))))
  (setq LS (car W)) (setq RS (cadr W)) (setq NS (caddr W))
  (cond ((eq '& (car A)) (setq A*
    (mapcan #'(lambda (u) 
      (mapcar #'(lambda (v)
        (list (l-le (car u) (car v)) (plus-on (cadr u) (cadr v)) )) RS )) LS))
	 (reduc (andify (append A* NS))))
	(t (setq A*
    (mapcan #'(lambda (u) 
      (mapcar #'(lambda (v)
        (list (l-le* (car u) (car v)) (plus-on (cadr u) (cadr v)))) RS )) LS))
         (cond ((eq 'or (car A)) (reduc (orify (append A* NS))))
	       (t                (reduc (append A* NS))) )) ))) ;clause case

(defun VE-the-variable (x &optional flag)     ;called by INEQUALITY-PROVE
  ;;Eliminate the (eligible) variable x from C if flag is t
  ;; or eliminate x from C, H, and YET, otherwise.

  (declare (special h c tl anc lt yet))

  (let (y y* A A* hy h*)

    (cond (flag (setq A c) (setq y* yet) (setq h* h))
	  (t     (setq y (remove-if-not #'(lambda (u) (occur x u)) YET))
			;y consists of those members of YET containing x 
		 (setq y* (set-difference YET y))

		 (setq y (remove-sub-here y))

		 (setq A (andify (mapcar #'(lambda (u) 
			      (list '-> (andify-h (cadr u)) (caddr u))) y)))

		 (setq hy (andify-h h))
		 (setq A (and-on (imp-on hy c) A))
		 (setq h* (list nil nil nil nil))))
    
    (cond ((setq A* (reduc* (ve x A)))
	   (format t "~%VE ~s   (The Variable ~s has been Eliminated.)~%" x x))
	  (t (return-from ve-the-variable nil)))

    (cond ((Inequality-prove h* A* (append tl (list 'VE x)) nil lt y*))
	  (t (imply h* A* (append tl (list 'VE x)) nil lt y*)) )))

(defun andify-h (h*)  ;used by VE-the-variable
    (and-on (car h*) (and-on (cadr h*) (and-on (caddr h*) (cadddr h*))))))

;;;****************  Examples  *********************************

(defvar c '(-> (-> (< 0 E)         ;Like LIM+   ??
        (& (< 0 (DS1 E))
           (-> (& (< (+ X (- (A))) (DS1 E))
                  (< (- (+ X (- (A)))) (DS1 E)))
               (& (< (+ ((F) X) (- ((F) (A)))) E)
                  (< (- (+ ((F) X) (- ((F) (A))))) E)))))
    (-> (& (< (D) ((F) (A))) (< ((F) (A)) (E)))
        (& (< A2 (A)) (< (A) B2)
           (-> (& (< A2 (XS1 B2 A2)) (< (XS1 B2 A2) B2))
               (& (< (D) ((F) (XS1 B2 A2))) (< ((F) (XS1 B2 A2)) (E))))))))


(defvar gt1 '(-> (& (<= a b) (<= b c) (<= c d)) (& (<= a c) (<= b d))))

(defvar e1 '(<= (+ X (- (A)))))

(DEFVAR e1* '(-> (NOT (& (<= (+ (A) (- X))) (< (+ X (- (B)))))) (C)))

(DEFVAR e2 '(& (<= (+ (A) (- X))) (<= (+ X (- (B))))))

(DEFVAR e3 '(-> (OR (<= (+ (A) (- X))) (< (+ X (- (B))))) (C)))

(DEFVAR e4 '(-> (-> (<= (+ X (- (A)))) (<= (+ X (- (B))))) (C)))

(DEFVAR e5 '(NOT (OR (<= (+ (A) (- X))) (< (+ X (- (B)))))))

(DEFVAR e6 '(-> (& (OR (<= (+ (A) (- X))) (< (+ X (- (B))))) (Q)) (C)))

(defvar e7 '(OR (& (<= (+ (A) (- X))) (< (+ X (- (B)))))
		(& (<= (+ (C) (- X))) (< (+ X (- (D)))))))

(defvar e8 '(-> (or (<= (a) x) (< x (b))) (& (<= (c) x) (< x (d)))))

(defvar e9 '(& (OR (& (<= (+ (A) (- X))) (< (+ X (- (B))))) (Q))
	       (<= (+ (C) (- X)))
	       (< (+ X (- (D))))))

(defvar e10 '(-> (OR (& (OR (<= (+ (A) (- X))) (< (+ X (- (B))))) (Q))
		     (<= (+ (C) (- X))) (< (+ X (- (D)))))
		 (C)))


